import os
import binascii
import struct
import hmac
import hashlib
from iotile.core.exceptions import NotFoundError
from .rootkey_auth_provider import RootKeyAuthProvider


class EnvAuthProvider(RootKeyAuthProvider):
    """Key provider implementation that search for keys in environment"""

    def __init__(self, args=None):
        if args is None:
            args = {}

        args['supported_keys'] = [self.UserKey]

        super().__init__(args)

    @classmethod
    def construct_var_name(cls, device_id):
        """Build name of environment variable used to store userkey

        Returns:
            str: variable name
        """
        if isinstance(device_id, str):
            var_name = "USER_KEY_{}".format(device_id)
        else:
            var_name = "USER_KEY_{0:08X}".format(device_id)
        return var_name

    def get_root_key(self, key_type, device_id):
        """Attempt to get a user key from an environment variable

        Args:
            key_type (int): see KnownKeyRoots
            device_id (int): uuid of the device

        Returns:
            bytes: the root key
        """
        self.verify_key(key_type)

        var_name = EnvAuthProvider.construct_var_name(device_id)

        if var_name not in os.environ:
            raise NotFoundError("No key could be found for devices", device_id=device_id,
                                expected_variable_name=var_name)

        key_var = os.environ[var_name]
        if len(key_var) != 64:
            raise NotFoundError("Key in variable is not the correct length, should be 64 hex characters",
                                device_id=device_id, key_value=key_var)
        try:
            key = binascii.unhexlify(key_var)
        except ValueError as exc:
            raise NotFoundError("Key in variable could not be decoded from hex", device_id=device_id,
                                key_value=key_var) from exc

        if len(key) != 32:
            raise NotFoundError("Key in variable is not the correct length, should be 64 hex characters",
                                device_id=device_id, key_value=key_var)

        return key



"""Hashing algorithms that we commonly use."""

from collections import namedtuple
from .crc import calculate_crc
from .sha import calculate_sha

HashAlgorithm = namedtuple("HashAlgorithm", ['calculate', 'algorithm'])

KNOWN_HASH_ALGORITHMS = {
    'crc32_0x104C11DB7': HashAlgorithm(calculate_crc, 0x104C11DB7),
    'sha256': HashAlgorithm(calculate_sha, 256)
}

__all__ = ['KNOWN_HASH_ALGORITHMS']



import os
import binascii
import struct
import hmac
import hashlib
from iotile.core.exceptions import NotFoundError
from .rootkey_auth_provider import RootKeyAuthProvider
from .env_auth_provider import EnvAuthProvider
from .auth_provider import AuthProvider
import getpass


class CliAuthProvider(RootKeyAuthProvider):
    """Key provider implementation that prompt user for password

       root key is derived using pbkdf2
    """
    def __init__(self, args=None):
        if args is None:
            args = {}

        args['supported_keys'] = [self.PasswordBasedKey]

        super().__init__(args)

    def get_root_key(self, key_type, device_id):
        """Prompt a user for the password and derive the root key from it

        Args:
            key_type (int): see KnownKeyRoots
            device_id (int): the uuid of the device

        Returns:
            bytes: the root key
        """
        self.verify_key(key_type)

        password = getpass.getpass("Please input the user password for the device {} :".format(device_id))
        userkey = AuthProvider.DeriveRebootKeyFromPassword(password)

        if device_id:
            answer = input("Would you like to save the user-key until the end of the current session? (y/n)")
            if answer and answer[0].lower() == 'y':
                variable_name = EnvAuthProvider.construct_var_name(device_id)
                os.environ[variable_name] = userkey.hex()

        return userkey



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# config.py

import os
import json


class ConfigFile:
    """A simple wrapper around a json config file with a local and global version
    the local version always overrides the global version.
    """

    def __init__(self, base_dir, name, require_local=False):
        localpath = os.path.join(base_dir, name +'.local.json')
        globalpath = os.path.join(base_dir, name + '.global.json')

        conf = {}
        localconf = None
        try:
            with open(globalpath, 'r') as f:
                conf = json.load(f)
        except IOError:
            pass #global file can not exist

        try:
            with open(localpath, 'r') as f:
                localconf = json.load(f)
        except IOError:
            if require_local:
                raise ValueError('Local settings file does not exist and is required: %s' % name + '.local.json')

        if localconf is not None:
            merge_dicts(conf, localconf)

        self.conf = conf

    def __getitem__(self, path):
        """Given a path specifying subdictionaries in this config dictionary,
        with each key separated by a /, return the value indicated.  It must
        exist.
        """

        keys = path.split('/')

        val = self.conf[keys[0]]

        for i in range(1, len(keys)):
            val = val[keys[i]]

        return val


# from http://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
def merge_dicts(a, b):
    "merges b into a"

    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key])
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]

    return a



"""Abstract base class for modern coroutine based DeviceAdapters.

All modern DeviceAdapter subclasses must implement this interface. This class
is the core interface by which any software built on top of CoreTools
interacts with an IOTile device.

High Level
==========

At a basic level, communication with an IOTile device through a DeviceAdapter
is connection oriented.  You connect to a device, then you open an interface
on that device.  You can then use that interface to interact with the device.
You close the interface when you are done and then you disconnect from the
device.


Implementation Details
----------------------

It is a very conscious choice to specify that :class:AbstractDeviceAdapter
must implement many of its methods as coroutines.  This is the product of much
experience implementing this functionality without coroutines and seeing the
complexity that it introduces into the internal implementation of a
DeviceAdapter.

Previous DeviceAdapter implementations used multithreading and callback based
interfaces that worked but proved difficult to implement, raising the barrier
to creating new DeviceAdapters to support additional communication protocols.


Events
======

Communication with a device through an AbstractDeviceAdapter is bidrectional
in that commands can flow from the user to the device and events can flow
back from the devie to the user without being the direct result of a command.

The way to receive events is to register an event monitor using
:meth:`AbstractDeviceAdapter.register_monitor`.  There are different kinds of
events that you can receive and your monitor includes filters based on the
event name and the source device.

Every event notification contains a tuple with three things: the connection
string of the device that sent the event, the name of the event and an event
object whose contents vary depending on the event name.  The supported events
are:

    report
        A report has been received from the device.

        The event object will be a subclass of
        :class:`iotile.core.hw.reports.IOTileReport`.
        This event is used to send reports from a device to the user via the
        streaming interface.  It can only be received if the streaming
        interface is open.

    connection
        Someone has connected to a device.

        The event object for this event will be ``None`` since there is not
        additional information available.

    trace
        Tracing data has been received from a device.

        The event object will be a ``bytes`` object containing a blob of
        tracing data.  Since tracing data may be fragmented as it travels from
        the device to you, there is no guarantee that you will receive it in
        the same atomic chunk sizes that the device sent.  In particular,
        trace data may be temporarily buffered inside some device adapters for
        performance reasons.

    device_seen
        A scan event has been received for a device.

        The event object will be a dictionary with certain mandatory keys and
        perhaps some additional keys whose meaning is determined by the device
        adapter itself.

    disconnection
        Someone has disconnected from a device.

        The event object will be a dictionary with at least a single key
        ``reason`` set that will be a string that describes why the disconnect
        occurred.  There will additionally be an ``expected`` key, which will
        be a boolean and set to True if the disconnetion event is expected
        because it was the result of a call to the
        :meth:`AbstractDeviceAdapter.disconnect`` method.

    broadcast
        A broadcast report has been received from a device.

        The event object will be a subclass of
        :class:`iotile.core.hw.reports.BroadcastReport`

    progress
        A progress update has been received from a device.

        Progress events help update the user on the sate of a long-running
        task.  This may be received either from a debug operation or a sending
        a script.

        The event object will be a dict with three keys set:

            operation
                The string name of the operation in progress: either ``script`` or
                ``debug``
            finished
                An integer specifying how many operation steps have finished.
            total
                An integer specifying how many total operation steps there are.
"""

import abc


class AbstractDeviceAdapter(abc.ABC):
    """The base class for all modern DeviceAdapters.

    This abstract class contains the interface that all DeviceAdapter classes
    must implement.  Any class implementing this interface can be used by
    CoreTools in order to interact with IOTile devices.  All hardware interaction
    in CoreTools is built on top of this foundation.

    Note that many of the methods in this class are specified to be
    implemented as coroutines. This is very important to allow for simple
    cooperative-multitasking around the typically long-running network
    communications required to perform operations on remote IOTile devices.
    """

    @abc.abstractmethod
    def can_connect(self):
        """Return whether this device adapter can accept another connection.

        Depending on the underlying hardware and transport protocols in use, a
        given DeviceAdapter may be able to maintain simultaneous connections
        to more than one device at a time.  In simple cases, this method is
        not necessary because a user can just try to connect to a device and
        have it fail if the adapter cannot accomodate another connection.

        However, in more complicated scenarios where there are multiple
        DeviceAdapters that could be used to connect to a device, it is useful
        to be able to filter them and find the best adapter that has a free
        connection slot available.

        Returns:
            bool: Whether one additional connection is possible
        """

    @abc.abstractmethod
    def get_config(self, name, default):
        """Get a configuration setting from this DeviceAdapter.

        This command can be used query features a DeviceAdapter supports
        before calling ``start`` to turn it on.  The list of keys that you can
        pass to ``name`` will be DeviceAdapter specific.

        This method is safe to call before ``start`` is called.  The
        DeviceAdapter will typically set many config values itself in its
        constructor that you can query with this method.

        Args:
            name (str): The name of the parameter to set
            default (object): The default value to return.  If no default is
                specified and the config parameter was never set an
                Exception will be raised.

        Raises:
            ArgumentError: The config does not exist and no default was specified.
        """

    @abc.abstractmethod
    def set_config(self, name, value):
        """Adjust a configuration setting on this DeviceAdapter.

        This command can be used to adjust how a DeviceAdapter will work
        before calling ``start`` to turn it on.  The list of keys that you can
        pass to ``name`` will be DeviceAdapter specific.

        This method is safe to call before ``start`` is called.  The
        DeviceAdapter will typically set many config values itself in its
        constructor that you are able to override in this method.

        Args:
            name (str): The name of the parameter to set
            value (object): The new value.
        """

    @abc.abstractmethod
    def register_monitor(self, devices, events, callback):
        """Register a callback when events happen.

        This method allows you to setup callbacks when events of interest
        happen inside of the device adapter.  You can register as many
        callbacks as you want and install filters for them to only receive a
        subset of events based on which device is involved or what kind of
        event it is.

        The supported event names are:

            report
                A report is received from the device.
            connection
                Someone has connected to a device.
            trace
                Tracing data has been received from a device.
            device_seen
                A scan event has been received for a device.
            disconnection
                Someone has disconnected from a device.
            broadcast
                A broadcast report has been received from a device.
            progress
                A progress update has been received from a device
                performing a long-running task.  This may be received
                either from a debug operation or a sending a script.

        The ``devices`` are a set of connection_string objects containing
        which devices you want to install this filter on.  All events are tied
        to a single device.  You can install a monitor to monitor a single
        device or all devices but keep in mind the volume of events you will
        get passed if installing a global monitor when there are many devices
        nearby.

        Coroutine callbacks are supported and will be awaited after each
        event.  Keep this in mind if your coroutine is long-running since it
        may block internal notification queues.

        All callbacks registered through this mechanism will be called in the
        following way::

            callback(conn_string, conn_id, event_name, event_object)

        The connection string will never be None but conn_id may be None for
        events that are associated with a device that does not have an active
        connection on this adapter.

        .. important:

            Once the adapter has started, this method may only be called from
            within the same event loop that is sending notification callbacks.
            All subclasses must ensure that this method works when called from
            within a notification callback itself.

        Args:
            devices (iterable of str): The devices, identified by their
                connection_string that you want to receive events for.  If you
                pass None here then all devices will be selected.
            events (iterable of str): The event names that you wish to register
                the callback for.
            callback (callable): The callback that should be invoked when an
                event happens.  If the callback is a coroutine function it
                will be awaited after each event.

        Raises:
            ArgumentError: The event name is unknown or the filters are invalid.

        Returns:
            object: A handle that can be used to adjust or remove the monitor.
        """

    @abc.abstractmethod
    def unique_conn_id(self):
        """Generate a new unique connection id.

        This method allows DeviceAdapters to tell their clients what they
        should pick for a connection id to ensure that it does not conflict
        with any current connection.  This can be imlemented as a simple
        incrementing counter, so clients must not mix calling unique_conn_id
        with setting their own connection ids.

        This method is very important when there could be multiple users
        connected to the same AbstractDeviceAdapter and they need to ensure
        that they don't accidentally pick the same connection id.

        Returns:
            int: A new, unique integer suitable for use as a conn_id.
        """

    @abc.abstractmethod
    def adjust_monitor(self, handle, action, devices, events):
        """Adjust a previously registered callback.

        This method allows you to adjust what events are listened for by a
        previously registered monitor or what devices are selected to receive
        events from.

        See :meth:`register_monitor` for the list of supported event names.

        If you remove a device or event that was not previously registered, no
        error is raised.  Similarly, if you add an event that was already
        registered, no error is raised.  The ``adjust_monitor`` method is
        idempotent.

        Args:
            handle (object): The handle to a previously registered monitor.
            action (str): One of ``add`` or ``remove``.
            devices (iterable of str): The devices that we wish to impact.
            events (iterable of str): The events that we wish to add or remove.

        Raises:
            ArgumentError: ``handle`` is not a registered monitor.
        """

    @abc.abstractmethod
    def remove_monitor(self, handle):
        """Remove a previously registered monitor.

        This method is idempotent so if the monitor does not exist, nothing
        occurs and no error is raised.

        Args:
            handle (object): The handle to a previously registered monitor.
        """

    @abc.abstractmethod
    async def probe(self):
        """Probe for visible devices connected to this DeviceAdapter.

        There are two ways by which a user could discover that they could
        connect to a device through the DeviceAdapter:

        1. The device could advertise itself periodically and the user could
           wait until they receive an ``on_found`` callback the next time the
           device adverises.

        2. There could be a mechanism to scan or probe for all connectable
           devices.  Not all DeviceAdapters are cabable of probing, but for
           those that are, it will ensure that ``on_found`` is called for any
           nearby connectable devices before returning.

        DeviceAdapter classes that can only find devices when explicitly
        searching for them should implement this method.  The method itself
        doesn't return anything but it should have a side effect that triggers
        ``on_found`` callbacks for any connectable devices.
        """

    @abc.abstractmethod
    async def connect(self, conn_id, connection_string):
        """Connect to a device.

        This coroutine should connect to a device using the passed connection
        string.  If the connection is succesful, this method returns nothing.
        If it fails, this method must raise a DeviceAdapterError exception
        that contains the reason for the failure.

        The ``conn_id`` parameter is an arbitrary unique integer passed in by
        the caller to identify the connection.  If connect() succeeds, then
        all future interaction with this device (while still connected) will
        reference ``conn_id`` as a pointer to this device.

        ``connection_string`` is an opaque str object that has meaning to
        this DeviceAdapter and lets it uniquely identify and connect to an
        IOTile Device.  These strings are always generated, originally,
        from the DeviceAdapter itself and sent to the user via ``on_found``
        callbacks when devices are observed nearby, or during a probe
        operation.

        The user calling this method does not necessarily know what the
        ``connection_string`` means that they are passing.  It is just an
        opaque token that this DeviceAdapter previously gave to the user to
        identify a connectable device.

        This means that before calling ``connect``, most users will need to
        have previously called ``probe`` to obtain a valid connection_string
        or registered an ``on_found`` event callback and waited a sufficient
        time for the callback to be called.

        Note that it is not required that only connection_strings returned
        from a prior ``on_found`` callback be passed to ``connect``.  If the
        connection_string has a semantic format that the user understands,
        e.g. it is a MAC address, then the user can try to connect to a device
        directly, without a discovery process, by constructing the appropriate
        connection_string manually.

        Args:
            conn_id (int): A unique identifier that will refer to this connection.
                The caller chooses this and it must be unique among all active
                connections but otherwise arbitrary.
            connection_string (str): A DeviceAdapter specific string that can be used
                to identify a connectable device.  The caller typically obtains this
                string by listening for ``on_found`` notifications or using the
                ``probe`` method.

        Raises:
            DeviceAdapterError: If the connection attempt was unsuccessful
        """

    @abc.abstractmethod
    async def disconnect(self, conn_id):
        """Disconnect from a connected device.

        This method will cleanly disconnect the given device and free any
        resources reserved for it on this DeviceAdapter.  It may only be
        called after a previous call to ``connect`` has succeeded and must be
        passed the same ``conn_id`` parameter in order to identify the
        appropriate device to disconnect from.

        Args:
            conn_id (int): The conn_id passed to a previously successful
                call to connect.

        Raises:
            DeviceAdapterError: The disconnection attempt was unsuccessful
        """

    @abc.abstractmethod
    async def open_interface(self, conn_id, interface):
        """Open an interface on a connected device.

        All interactions with a device through a DeviceAdapter happen
        on an interface.  There are different kinds of interfaces and
        they have different operations that they support.  The five
        kinds of interfaces are:

            rpc
                This interface allows you to send a synchronous command
                to the device and get a response like calling a function
                or invoking a shell command.

            script
                This interface allows you to send a multicommand script
                to the device at high speed, which will be stored and
                may be triggered to run at a later point using an rpc.

            streaming
                This interface is unidirectional from the device to the
                user and allows it to send universally understandable
                data reports that contain timestamped sensor or other
                readings.

            tracing
                This is a unidirectional socket-like interface from the
                device to the user that can be used to send arbitary
                binary data.

            debug
                This is a debugging interface that can be used to send
                out-of-band commands to the device or associated debug
                circuitry around it such as manually reflashing its
                firmware, inspecting its memory etc.

        Not all devices implement all 5 interfaces, nor do all DeviceAdapters.
        You must open an interface before you are able to use it (and you
        should close it when you done).  What actual physical actions are
        performed when you "open an interface" are unspecified and vary widely
        from DeviceAdapter to DeviceAdapter based on the needs of the
        underlying communication protocol.

        You must be connected to a device before you can open an interface on
        it.

        Args:
            interface (str): The interface to open.  This must be one of the
                following strings: ``rpc``, ``script``, ``streaming``, ``tracing``,
                ``debug``.
            conn_id (int): The connection id used in a previous successful call
                to ``connect``.

        Raises:
            DeviceAdapterError: The interface could not be opened.
        """

    @abc.abstractmethod
    async def close_interface(self, conn_id, interface):
        """Close an interface on a device.

        The interface parameter must be one of (rpc, script, streaming,
        tracing, debug).  You must hvae previously opened the interface
        using ``open_interface`` for this call to succeed.

        Args:
            interface (string): The interface to close
            conn_id (int): The connection id used in a previous successful call
                to ``connect``.

        Raises:
            DeviceAdapterError: The interface could not be closed.
        """

    @abc.abstractmethod
    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Send an RPC to a device.

        This method is the only way to interact with the RPC interface of a
        device.  You specify two numbers: ``address`` and ``rpc_id`` that
        uniquely identify a remotely callable procedure inside the device and
        pass a payload containing the marshalled command arguments.

        The procedure is invoked and its return value is marshalled and sent
        back to you as a binary blob.

        You can specify a timeout for the maximum amount of time you expect
        the RPC to take so that DeviceAdapter can set an appropriate internal
        timeout to protect against communication failures while not failing
        too soon if the RPC is known to be long-running.

        The exact meaning and size of ``address`` and ``rpc_id`` are technically
        specified by the device itself but most DeviceAdapters and devices
        enforce the following limits:

        - address: 1 bytes
        - rpc_id: 2 bytes
        - payload: up to 20 bytes
        - response: up to 20 bytes

        The combination of ``address`` and ``rpc_id`` must uniquely identify a
        single RPC implementation on the device.  It is not required that
        ``rpc_id`` itself is unique.

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the numerical id of the RPC we want to call
            payload (bytes): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute

        Returns:
            bytes: The response payload returned by the RPC.

        Raises:
            TileNotFoundError: The destination tile address does not exist
            RPCNotFoundError: The rpc_id does not exist on the given tile
            RPCErrorCode: The RPC implementation wishes to fail with a
                non-zero status code.
            RPCInvalidIDError: The rpc_id is too large to fit in 16-bits.
            TileBusyError: The tile was busy and could not respond to the RPC.
            Exception: The rpc raised an exception during processing.
            DeviceAdapterError: If there is a hardware or communication issue
                invoking the RPC.
        """

    @abc.abstractmethod
    async def debug(self, conn_id, name, cmd_args):
        """Send a debug command to a device.

        The command name and arguments are passed to the underlying device
        adapter and interpreted there.  If the command is long running,
        you may receive progress events (if you have a registered event monitor).

        Most DeviceAdapter classes do not support this command or the debug
        interface since it typically requires a special hardware connection to
        the device.  However, for those that do, this method provides helpful
        debug related features.

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            name (str): The name of the debug command we want to invoke
            cmd_args (dict): Any arguments that we want to send with this command.
        """

    @abc.abstractmethod
    async def send_script(self, conn_id, data):
        """Send a a script to a device.

        Scripts are like binary files transferred to a device.  They are
        typically used to perform atomic device updates.

        The device must have previously been connected to and you must have
        opened the script interface.  This method will robustly transfer a binary
        script blob to the device via its script interface and may inform you
        of its progress using progress events.

        Args:
            conn_id (int): A unique identifier specifying the connection
            data (bytes): The script to send to the device
        """

    @abc.abstractmethod
    async def start(self):
        """Start this AsyncDeviceAdapter.

        This method may only be called once in a DeviceAdapter's life and it
        will start any background activity that is needed and prepare the
        device adapter to function.  No other methods should be called before
        start() is called, except for ``get_config`` and ``set_config`` in
        order to perform any necessary adjustments.
        """

    @abc.abstractmethod
    async def stop(self):
        """Stop this AsyncDeviceAdapter.

        This method may only be called once in an AsyncDeviceAdapter's life
        and it will stop all background activity and free all resources
        associated with the DeviceAdapter.  No other calls should be made to
        the device adapter after stop() is called.
        """



"""A convenience subclass for simple virtual devices.

These devices do not include support for tiles or complex modularity but
instead just queue a few streams and/or traces when a user connects to them and
support running periodic background tasks using coroutine based workers that
are triggered on a schedule.

They are useful for writing simple, repeatable tests.
"""

import inspect
from iotile.core.utilities import SharedLoop
from ..exceptions import RPCInvalidIDError, RPCNotFoundError, TileNotFoundError
from .virtualdevice_base import BaseVirtualDevice
from .periodic_worker import _PeriodicWorkerMixin
from .common_types import rpc, RPCDispatcher


class SimpleVirtualDevice(BaseVirtualDevice, _PeriodicWorkerMixin):
    """A simple virtual device with a "fake" tile at address 8.

    This class implements the required controller status RPC that allows
    matching it with a proxy object.  You can define periodic worker functions
    that add simple interactivity using the :meth:`start_worker` method.

    If you want to add extra RPCs, you can do so by decorating your rpc
    implementation with an ``@rpc`` decorator or by adding them dynamically
    using :meth:`register_rpc`.

    The default implementation will respond to RPC 8:0004 with a fixed tile
    name for the tile at address 8 based on the ``name`` you pass in to the
    __init__ function.  This allows for the creation of matched proxy objects
    that correspond with your simple device.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        name (str): The 6 byte name that should be returned when anyone asks
            for the controller's name of this IOTile device using an RPC
        loop (BackgroundEventLoop): The loop we should use for running background
            tasks. Defaults to the global SharedLoop.
            *This must be passed as a keyword agument.*
    """

    __NO_EXTENSION__ = True

    def __init__(self, iotile_id, name, *, loop=SharedLoop):
        BaseVirtualDevice.__init__(self, iotile_id, loop=loop)
        _PeriodicWorkerMixin.__init__(self)

        self.name = name.encode('utf-8')
        self.reports = []
        self.traces = []
        self.script = bytearray()

        # Iterate through all of our member functions and see the ones that are
        # RPCS and add them to the RPC handler table
        self._rpc_overlays = {}
        for _name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.register_rpc(value.rpc_addr, value.rpc_id, value)

    def register_rpc(self, address, rpc_id, func):
        """Register a single RPC handler with the given info.

        This function can be used to directly register individual RPCs,
        rather than delegating all RPCs at a given address to a virtual
        Tile.

        If calls to this function are mixed with calls to add_tile for
        the same address, these RPCs will take precedence over what is
        defined in the tiles.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            func (callable): The function that should be called to handle the
                RPC.  func is called as func(payload) and must return a single
                string object of up to 20 bytes with its response
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self._rpc_overlays:
            self._rpc_overlays[address] = RPCDispatcher()

        self._rpc_overlays[address].add_rpc(rpc_id, func)

    async def async_rpc(self, address, rpc_id, payload=b""):
        """Call an RPC by its address and ID.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            bytes: The response payload from the RPC
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self._rpc_overlays:
            raise TileNotFoundError("Unknown tile address, no registered handler", address=address)

        overlay = self._rpc_overlays.get(address, None)

        if overlay is not None and overlay.has_rpc(rpc_id):
            resp = overlay.call_rpc(rpc_id, payload)
            if inspect.isawaitable(resp):
                resp = await resp

            return resp

        raise RPCNotFoundError("Could not find RPC 0x%X at address %d" % (rpc_id, address))


    @rpc(8, 0x0004, "", "H6sBBBB")
    def status(self):
        """RPC returning a tile name to match against an installed proxy."""

        status = (1 << 1) | (1 << 0) # Configured and running

        return [0xFFFF, self.name, 1, 0, 0, status]

    def push_script_chunk(self, chunk):
        """Called when someone pushes a new bit of a TRUB script to this device

        Args:
            chunk (str): a buffer with the next bit of script to append
        """

        self.script += bytearray(chunk)

    def _open_streaming_interface(self):
        """Handler called when the streaming interface is opened.

        If a list of reports is returned, those reports are streamed out of
        the device immediately.
        """

        return self.reports

    def _open_tracing_interface(self):
        """Handler called when the tracing interface is opened.

        If a list of bytes-like object is returned, they will be traced out of
        the tracing interface immediately.
        """

        return self.traces

    def start(self, channel):
        """Start running this virtual device including any necessary worker threads.

        Args:
            channel (AbstractAsyncDeviceChannel): channel to use to push asynchronous
                events to a connected client.
        """

        super(SimpleVirtualDevice, self).start(channel)
        self.start_workers()

    def stop(self):
        """Stop running this virtual device including any worker threads."""

        self.stop_workers()
        super(SimpleVirtualDevice, self).stop()



"""A VirtualDevice composed of various tiles.

The device handles RPCs by dispatching them to tiles configured using a config dictionary.
"""
from iotile.core.hw.virtual import tile_rpc
from .virtualdevice_standard import StandardVirtualDevice
from .virtualtile import VirtualTile


class TileBasedVirtualDevice(StandardVirtualDevice):
    """A VirtualDevice composed of one or more tiles.

    Args:
        args (dict): A dictionary that lists the tiles that
            should be loaded to create this virtual device.
        override_controller (bool): Controls whether the virtual controller
            gets added or not. Defaults to false.
    """

    def __init__(self, args):
        iotile_id = args.get('iotile_id')

        override_controller = args.get('override_controller', False)

        con_args = args

        if isinstance(iotile_id, str):
            iotile_id = int(iotile_id, 16)

        super(TileBasedVirtualDevice, self).__init__(iotile_id)

        for desc in args.get('tiles', []):
            name = desc['name']
            address = desc['address']

            args = desc.get('args', {})

            tile_type = VirtualTile.FindByName(name)
            tile = tile_type(address, args, device=self)

            self.add_tile(address, tile)

        # Add a controller tile
        if not override_controller:
            tile = VirtualController(8, con_args, device=self)
            self.add_tile(8, tile)
            tile.tiles = self._tiles
            tile.prep_count()

    def start(self, channel):
        """Start running this virtual device including any necessary worker threads.

        Args:
            channel (IOTilePushChannel): the channel with a stream and trace
                routine for streaming and tracing data through a VirtualInterface
        """

        super(TileBasedVirtualDevice, self).start(channel)

        for tile in self._tiles.values():
            tile.start(channel=channel)

    def stop(self):
        """Stop running this virtual device including any worker threads."""

        for tile in self._tiles.values():
            tile.stop()

        super(TileBasedVirtualDevice, self).stop()


class VirtualController(VirtualTile):
    """A tile based virtual controller. Contains a very limited subset of controller RPCs currently.

    When using it, you should force your program to see it as an 'NRF52 ' proxy, because vircon is
    not an actual proxy object (there are no unique RPCs here outside of the base controller spec).
    The syntax for doing this is something like:

    get 8 -f 'NRF52 '
    (assuming you are connected to the device already).

    To use the RPCs, you should use the test_interface calls for get_info and set_version.
    """

    def __init__(self, address, config, device):

        self.config = config
        super(VirtualController, self).__init__(address, 'vircon')

        self.tile_list = []

    def prep_count(self):
        """Put the actual tile indices in to a list that complies with the count_tiles and describe_tile API"""
        for tile in self.tiles:
            self.tile_list.append(tile)

    @tile_rpc(0x0004, "", "H6sBBBB")
    def status(self):
        """Return a dummy status"""
        return [0xFFFF, "vircon", 1, 0, 0, 1]

    @tile_rpc(0x0002, "", "10s")
    def hardware_version(self):
        """Return the hardware identification string."""
        return [b'virtualcon']

    @tile_rpc(0x2a01, "", "H")
    def count_tiles(self):
        """Count the number of registered tiles including the controller"""
        return [len(self.tile_list)]

    @tile_rpc(0x2a02, "H", "3B6s6BBL")
    def describe_tile(self, index):
        """Describes a tile at the given index

        The information returned is the exact same data as what REGISTER_TILE
        includes as its argument.

        Args:
        - uint16_t: The index of the tile you wish to describe.  Must be less than
            what COUNT_TILES returns.

        Returns:
        - uint8_t: a numerical hardware type identifer for the processor the tile is
            running on.
        - two bytes: (major, minor) API level of the tile fwutive running on this
            tile.  The api version is encoded with the major version as a byte
            followed by the minor version.
        - 6-character string: The tile firmware identifier that is used to match it
            with the correct proxy module by HardwareManager.
        - 3 bytes: (major, minor, patch) version identifier of the application
            firmware running on this tile.
        - 3 bytes: (major, minor, patch) version identifier of the tile executive
            running on this tile.
        - uint8_t: The slot number that refers to the physical location of this tile
            on the board.  It is used to assign the tile a fixed address on the
            TileBus.
        - uint32_t: A unique identifier for this tile, if the tile supports unique
            identifiers.
        """
        tile = self.tiles[self.tile_list[index]]
        return (0, 0, 0, tile.name, 0, 0, 0, 0, 0, 0, self.tile_list[index], 0)



""" Interface that provides access to keys"""

import hashlib
import struct
import hmac
from iotile.core.exceptions import NotFoundError

class AuthProvider:
    """Base class for all objects that provide a way to obtain a key

    Args:
        args (dict): An optional dictionary of AuthProvider specific config information
            field "supported_keys" specify wich keys the AuthProvider supports
    """

    ReportKeyMagic = 0x00000002

    NoKey = 0
    NullKey = 1
    UserKey = 2
    DeviceKey = 3
    PasswordBasedKey = 4

    KnownKeyRoots = {
        NoKey: 'no_key',
        NullKey: 'null_key',
        UserKey: 'user_key',
        DeviceKey: 'device_key',
        PasswordBasedKey: 'pasword_based_userkey'
    }

    def __init__(self, args=None):
        if args is None:
            args = {}

        self.args = args
        self.supported_keys = self.args.get("supported_keys", [])

    def verify_key(self, root_key_type):
        """Verify that the root key type is known to us.

        Args:
            root_key_type: requested key type

        Raises:
            NotFoundError: If the key type is not known.
        """

        if root_key_type not in self.KnownKeyRoots:
            raise NotFoundError("Unknown key type", key=root_key_type)

        if root_key_type not in self.supported_keys:
            raise NotFoundError("Not supported key type", key=root_key_type)

    @classmethod
    def DeriveRebootKeyFromPassword(cls, password):
        """Derive the root key from the user password
        TODO hashlib.pbkdf2_hmac arguments needs to be revised,
            current values are not proved to be secure

        Args:
            password (str): user password

        Returns:
            bytes: derived key
        """
        return hashlib.pbkdf2_hmac('sha256', password.encode('utf-8'), b'salt', 100000)

    @classmethod
    def DeriveReportKey(cls, root_key, report_id, sent_timestamp):
        """Derive a standard one time use report signing key.

        The standard method is HMAC-SHA256(root_key, MAGIC_NUMBER || report_id || sent_timestamp)
        where MAGIC_NUMBER is 0x00000002 and all integers are in little endian.

        Args:
            root_key (bytearray): 16 bytes with the key
            report_id (int): id in report
            sent_timestamp (int): timestamp in report

        Returns:
            bytearray: 32 bytes containing digest of id and timestamp
        """

        signed_data = struct.pack("<LLL", AuthProvider.ReportKeyMagic, report_id, sent_timestamp)

        hmac_calc = hmac.new(root_key, signed_data, hashlib.sha256)
        return bytearray(hmac_calc.digest())


    @classmethod
    def DeriveRebootKey(cls, root_key, key_purpose, reboot_counter):
        """Derive a key generated on reboot

        HMAC SHA-256-128(Root Key, key purpose (uint32_t) || reboot counter (uint32_t))

        Args:
            root_key (bytearray): 16 bytes key
            key_purpose (int): value to differentiate different keys,
                for example with different permissions. Not used at the moment
            reboot_counter (int): received in ble adv packet

        Returns:
            bytearray: 16 bytes containing digest of key_purpose and reboot_counter
        """
        signed_data = struct.pack("<LL", key_purpose, reboot_counter)

        hmac_calc = hmac.new(root_key, signed_data, hashlib.sha256)
        return bytearray(hmac_calc.digest()[:16])


    @classmethod
    def DeriveRotatedKey(cls, reboot_key, current_timestamp, rotation_interval_power):
        """Derive an ephemeral key every 2^rotation_interval_power seconds,
        The same key will be return if timeout of rotation is not elapsed

        AES-128-ECB(Reboot Key, current_timestamp with low X bits masked to 0)

        Args:
            reboot_key (bytearray): 16 bytes key
            current_timestamp (int): current time
            rotation_interval_power (int): X in 2^X

        Returns:
            bytearray: the rotated key
        """
        timestamp = current_timestamp & (~(2 ** rotation_interval_power - 1))

        try:
            from Crypto.Cipher import AES
        except ImportError as error:
            raise NotFoundError("Cryptographic library is not available") from error

        cipher = AES.new(reboot_key, AES.MODE_ECB)
        msg = struct.pack("<LLLL", timestamp, 0, 0, 0)

        return cipher.encrypt(msg)


    def get_serialized_key(self, key_type, device_id, **key_info):
        """Get a serialized key such for signing a streamer report.

        These keys are designed to only be used once and only provide
            access to the object of key_type with the given serial_number

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device
            key_info (dict): data required for key generation.
                It may be report_id and sent_timestamp

        Returns:
            bytearray: the key
        """
        raise NotImplementedError()


    def get_rotated_key(self, key_type, device_id, **rotation_info):
        """Get a key that is only valid for a limit period of time.

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device
            key_info (dict): data that describes the conditions when
                the key is rotated. For example, for a broadcast report key
                it may be the reboot counter of the
                device, the current uptime and the rotation interval of the key.

        Returns:
            bytearray: the rotated key
        """
        raise NotImplementedError()


    def get_root_key(self, key_type, device_id):
        """Get the actual root key of a given type for a device.

        Few AuthProvider classes should actually implement this method
        since it's safer to only handle derived keys that give specific
        abilities.  For AuthProviders that have access to the root key
            though, the other methods can be provided via a mixin from this
            root method.

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device

        Returns:
            bytearray: the root key
        """

        raise NotImplementedError()


    def get_device_access_key(self, key_type, device_id, scope):
        """Future method for scoped access tokens to device.

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device
            scope (int): permissions that will be granted with this key

        Returns:
            bytearray: the key

        """
        raise NotImplementedError()



"""Message driven asynchronous state machines.

This module allows you to define a complex operation that blocks at certain
points to wait for one or more messages to be received.  There can be
many operations running in parallel and multiple operations can be triggered
by the same message.  Messages are just dictionaries and the matching can
be according to any keys in the messages.

The OperationManager class is designed to provide a friendly API on which to
build complex, asynchronous network or hardware operations.

The prototypical use case is dealing with messages received from bluetooth low
energy hardware. There is a single channel over which all messages are
received but they are logically multiplexed to multiple different parallel
connections.

In each of those connections you could have multiple parallel operations
happening (such as sending data to one characteristics while reading data from
another).  Each of those operations could be a multi-step non-atomic process
where the step transitions consist of waiting for the receipt of one or more
messages.

Implementing these operations usually involves complex callback or
multithreading based code.  OperationManager provides a tailored API for
allowing a coroutine to block until a specific message is received (or one of
several messages are received).

The goal is to allow the creation of simple coroutines that can safely share
access to to the underlying multiplexed channel and clearly specify what
messages they are waiting for to continue their operation.
"""

import asyncio
import inspect
import logging
from collections import deque
from iotile.core.exceptions import ArgumentError, InternalError
from . import event_loop


class MessageSpec:
    __slots__ = ['fields']

    def __init__(self, **kwargs):
        self.fields = kwargs


class OperationManager:
    """Manage complex operations blocking on specific network messages.

    This utility class is designed to provide a friendly API on which to build
    complex, asynchronous network or hardware operations.  The prototypical
    use case is dealing with messages received from bluetooth low energy
    hardware.

    You pass it messages using the process_message() API, where a message can
    either be an object with properties or a dictionary.  Each message is then
    compared to see if anyone is waiting for that message and if so, the
    waiter is unblocked or a callback is called.

    The goal is to be able to write coroutines that cleanly block until a
    specific message is received or timeout safely.

    You can block on a message by calling OperationManager.wait_for() with a
    specification of the message you are looking for.  If you just want a
    callback every time a certain kind of message is received, use
    OperationManager.every_match().

    Args:
        loop (BackgroundEventLoop): The background event loop we should
            use to perform all callbacks and blocking waits.  If not specified,
            the global, shared event loop is used.
    """

    _LEAF = object()

    def __init__(self, loop=event_loop.SharedLoop):
        if not isinstance(loop, event_loop.BackgroundEventLoop):
            raise ArgumentError("loop must be a BackgroundEventLoop, was {}".format(loop))

        self._waiters = {}
        self._should_pause = set()
        self._loop = loop
        self._logger = logging.getLogger(__name__)
        self._pause_count = 0
        self._messages = deque()
        self._dispatch_lock = loop.create_lock()
        self._process_pending = False

    def _add_waiter(self, spec, responder=None, pause=False):

        if responder is not None and pause is True:
            raise ArgumentError("You can only pause after waiting for future based waiters")

        loc = self._waiters
        for key, value in sorted(spec.fields.items()):
            if key not in loc:
                loc[key] = {}

            loc = loc[key]
            if value not in loc:
                loc[value] = {}

            loc = loc[value]

        if OperationManager._LEAF not in loc:
            loc[OperationManager._LEAF] = set()

        if responder is None:
            responder = self._loop.create_future()

        loc[OperationManager._LEAF].add(responder)

        if pause:
            self._should_pause.add(responder)

        return responder

    def _remove_waiter(self, spec, future):
        loc = self._waiters

        parents = []

        for key, value in sorted(spec.fields.items()):
            parents.append((loc, key))

            loc = loc.get(key)
            if loc is None:
                return

            parents.append((loc, value))

            loc = loc.get(value)
            if loc is None:
                return

        futures = loc.get(OperationManager._LEAF)
        if futures is None or future not in futures:
            return

        futures.remove(future)
        if len(futures) > 0:
            return

        del loc[OperationManager._LEAF]

        for parent, key in reversed(parents):
            if len(parent[key]) is not None:
                return

            del parent[key]

        if future in self._should_pause:
            self._should_pause.remove(future)

    def waiters(self, path=None):
        """Iterate over all waiters.

        This method will return the waiters in unspecified order
        including the future or callback object that will be invoked
        and a list containing the keys/value that are being matched.

        Yields:
            list, future or callable
        """

        context = self._waiters

        if path is None:
            path = []

        for key in path:
            context = context[key]

        if self._LEAF in context:
            for future in context[self._LEAF]:
                yield (path, future)

        for key in context:
            if key is self._LEAF:
                continue

            yield from self.waiters(path=path + [key])

    def every_match(self, callback, **kwargs):
        """Invoke callback every time a matching message is received.

        The callback will be invoked directly inside process_message so that
        you can guarantee that it has been called by the time process_message
        has returned.

        The callback can be removed by a call to remove_waiter(), passing the
        handle object returned by this call to identify it.

        Args:
            callback (callable): A callable function that will be called as
                callback(message) whenever a matching message is received.

        Returns:
            object: An opaque handle that can be passed to remove_waiter().

            This handle is the only way to remove this callback if you no
            longer want it to be called.
        """

        if len(kwargs) == 0:
            raise ArgumentError("You must specify at least one message field to wait on")

        spec = MessageSpec(**kwargs)
        responder = self._add_waiter(spec, callback)

        return (spec, responder)

    def _add_temporary_waiter(self, spec):
        future = self._add_waiter(spec)
        future.add_done_callback(lambda x: self._remove_waiter(spec, future))

        return future

    def remove_waiter(self, waiter_handle):
        """Remove a message callback.

        This call will remove a callback previously registered using
        every_match.

        Args:
            waiter_handle (object): The opaque handle returned by the
                previous call to every_match().
        """

        spec, waiter = waiter_handle
        self._remove_waiter(spec, waiter)

    def clear(self):
        """Clear all waiters.

        This method will remove any current scheduled waiter with an
        asyncio.CancelledError exception.
        """

        for _, waiter in self.waiters():
            if isinstance(waiter, asyncio.Future) and not waiter.done():
                waiter.set_exception(asyncio.CancelledError())

        self._waiters = {}

    def wait_for(self, timeout=None, pause=False, unpause=False, **kwargs):
        """Wait for a specific matching message or timeout.

        You specify the message by passing name=value keyword arguments to
        this method.  The first message received after this function has been
        called that has all of the given keys with the given values will be
        returned when this function is awaited.

        If no matching message is received within the specified timeout (if
        given), then asyncio.TimeoutError will be raised.

        This function only matches a single message and removes itself once
        the message is seen or the timeout expires.

        Args:
            timeout (float): Optional timeout, defaults to None for no timeout.
            pause (bool): Pause the delivery of messages after receiving the message
                that this method is waiting for.
            unpause (bool): Unpause message delivery before waiting for messages.
            **kwargs: Keys to match in the message with their corresponding values.
                You must pass at least one keyword argument so there is something
                to look for.

        Returns:
            awaitable: The response
        """

        if len(kwargs) == 0:
            raise ArgumentError("You must specify at least one message field to wait on")

        spec = MessageSpec(**kwargs)
        future = self._add_waiter(spec, pause=pause)
        future.add_done_callback(lambda x: self._remove_waiter(spec, future))

        if unpause:
            self.unpause()

        return asyncio.wait_for(future, timeout=timeout)

    async def gather_until(self, gather_specs, until, timeout=None, *, pause=False, unpause=False):
        """Gather all matching messages until a message matching until is given or a timeout.

        This method allows you to accumulate messages matching the ``MessageSpec`` objects in
        ``gather_specs`` until a message matching the MessageSpec ``until`` is seen or a
        timeout occurs.

        Args:
            gather_specs (Iterable(MessageSpec)): The message specs that should be accumulated
            until (MessageSpec): The message spec that signals the end of gathering.  This will
                not be returend unless you also include the same message spec in ``gather_specs``.
            timeout (float): The maximum amount of time to wait for a message matching ``until``
                before giving up.  This is in seconds.
            pause (bool): Pause the delivery of messages after receiving the message
                that this method is waiting for.
            unpause (bool): Unpause message delivery before waiting for messages.

        Returns:
            list(object): A list of the gathered messages.

        Raises:
            asyncio.TimeoutError: The timeout expired before a message matching ``until`` was seen.
        """

        accum = []

        def _accum_message(message):
            accum.append(message)

        handles = []
        for spec in gather_specs:
            resp = self._add_waiter(spec, _accum_message)
            handles.append((spec, resp))

        future = self._add_waiter(until, pause=pause)
        future.add_done_callback(lambda x: self._remove_waiter(until, future))

        if unpause:
            self.unpause()

        try:
            await asyncio.wait_for(future, timeout=timeout)
        finally:
            for spec, resp in handles:
                self._remove_waiter(spec, resp)

        return accum

    async def gather_count(self, gather_specs, count, timeout=None, *, pause=False, unpause=False):
        """Gather all matching messages until a fixed number have been received.

        This method waits until an exact number of messages matching any of the specs in ``gather_specs``
        is received or a timeout occurs.

        Args:
            Args:
            gather_specs (Iterable(MessageSpec)): The message specs that should be accumulated
            count (int): The number of matching messages to wait for.
            timeout (float): The maximum amount of time to wait for a message matching ``until``
                before giving up.  This is in seconds.
            pause (bool): Pause the delivery of messages after receiving the message
                that this method is waiting for.
            unpause (bool): Unpause message delivery before waiting for messages.

        Returns:
            list(object): A list of the messages that were gathered.

            If this method returns normally the list will always have exactly ``count`` entries.

        Raises:
            asyncio.TimeoutError: The timeout expired before ``count`` matching messages were seen.
        """

        accum = []
        done = self._loop.create_event()

        def _accum_message(message):
            accum.append(message)

            if len(accum) == count:
                done.set()
                if pause:
                    self.pause()

        handles = []
        for spec in gather_specs:
            resp = self._add_waiter(spec, _accum_message)
            handles.append((spec, resp))

        if unpause:
            self.unpause()

        try:
            await asyncio.wait_for(done.wait(), timeout=timeout)
        except asyncio.TimeoutError as err:
            raise asyncio.TimeoutError("Timeout waiting for gathering fixed number of messages (%d of %d received)"
                                       " pause_count=%d"
                                       % (len(accum), count, self._pause_count)) from err
        finally:
            for spec, resp in handles:
                self._remove_waiter(spec, resp)

        return accum

    async def process_message(self, message, wait=True):
        """Process a message to see if it wakes any waiters.

        This will check waiters registered to see if they match the given
        message.  If so, they are awoken and passed the message.  All matching
        waiters will be woken.

        This method returns False if the message matched no waiters so it was
        ignored.

        All waiters for a given message are run in parallel with no guaranteed
        ordering.  However, it is guaranteed that all coroutines that were
        triggered by this message have run until completion by the time this
        method returns.

        If any waiters were configured to pause processing after they matched
        a message, this OperationManager will be paused when this method
        returns.

        Args:
            message (dict or object): The message that we should process
            wait (bool): Whether to wait for the message handlers to finish
                running. Defaults to True.

        Returns:
            bool: True if at least one waiter matched, otherwise False.
        """

        to_check = deque([self._waiters])
        ignored = True

        processors = list()

        while len(to_check) > 0:
            context = to_check.popleft()

            waiters = context.get(OperationManager._LEAF, [])
            for waiter in waiters:
                if isinstance(waiter, asyncio.Future):
                    if waiter in self._should_pause:
                        self.pause()

                    waiter.set_result(message)
                else:
                    proc = _launch(waiter, message, wait)
                    if proc is not None:
                        processors.append(proc)

                ignored = False

            for key in context:
                if key is OperationManager._LEAF:
                    continue

                message_val = _get_key(message, key)
                if message_val is _MISSING:
                    continue

                next_level = context[key]
                if message_val in next_level:
                    to_check.append(next_level[message_val])

        if len(processors) > 0 and wait:
            results = await asyncio.gather(*processors, return_exceptions=True)
            for proc, result in zip(processors, results):
                if isinstance(result, Exception) and not isinstance(result, asyncio.CancelledError):
                    self._logger.error("Error running processor %s: %s", proc, result)

        return not ignored

    async def process_all(self):
        """Process all messages that have been queued.

        You can queue messages using queue_message_threadsafe.  This method
        will drain the queue and process them all one at a time.  This method
        is automatically called as needed by queue_message_threadsafe and
        pause/unpause so it should not ever need to be called manually outside
        of testing scenarios.

        Only a single call to process all can be running at a time to ensure
        that messages are always dispatched in order without race conditions.
        """

        self._process_pending = False

        async with self._dispatch_lock:
            while len(self._messages) > 0:
                if self._pause_count > 0:
                    return

                current = self._messages.popleft()
                await self.process_message(current)

    def queue_message_threadsafe(self, message):
        """Queue a message for processing."""

        self._messages.append(message)
        if not self._process_pending:
            self._process_pending = True

            if self._pause_count == 0:
                self._loop.launch_coroutine(self.process_all)

    def pause(self):
        """Request to pause the dispatch of messages.

        This method increases an internal pause counter.  As long as the pause
        count is greater than 0, no messages will be dispatched through
        process_all().  This is used in combination with routines like gather_until()
        to make sure messages don't get processed in between when one operation finishes
        and you setup the next operation.
        """

        self._pause_count += 1

    def unpause(self):
        """Undo a previous call to pause().

        This method decreases the internal pause request counter.  When the counter
        reaches 0, message processing is allowed again.
        """

        if self._pause_count == 0:
            raise InternalError("Mismatched usage of pause/unpause, unpause called when not paused")

        self._pause_count -= 1
        if self._pause_count == 0 and len(self._messages) > 0:
            self._loop.launch_coroutine(self.process_all)


_MISSING = object()

def _get_key(obj, key, default=_MISSING):
    if isinstance(obj, dict):
        return obj.get(key, default)

    if hasattr(obj, key):
        return getattr(obj, key)

    return default


def _launch(coroutine_func, message, wait):
    result = coroutine_func(message)
    if inspect.isawaitable(result):
        if wait:
            return result

        asyncio.ensure_future(result)
        return None

    return None



"""Shared decorators and exceptions used in virtual tiles and devices."""

import struct
from collections import namedtuple
import binascii
import inspect
from ..exceptions import (RPCNotFoundError, RPCInvalidArgumentsError,
                          RPCInvalidReturnValueError, RPCInvalidIDError,
                          TileNotFoundError, RPCErrorCode,
                          BusyRPCResponse)


RPCDeclaration = namedtuple("RPCDeclaration", ["rpc_id", "arg_format", "resp_format"])


def _create_argcode(code, arg_bytes):
    if not code.endswith('V'):
        return "<" + code

    code = code[:-1]
    fixed_size = struct.calcsize("<" + code)
    var_size = len(arg_bytes) - fixed_size

    if var_size < 0:
        raise RPCInvalidArgumentsError("Argument was too small for variable size argument value", arg_format=code,
                                       minimum_size=fixed_size, actual_size=len(arg_bytes),
                                       payload=binascii.hexlify(arg_bytes))

    return "<" + code + "%ds" % var_size


def _create_respcode(code, resp):
    if not code.endswith('V'):
        return "<" + code

    code = code[:-1]

    final_length = len(resp[-1])
    fixed_size = struct.calcsize("<" + code)

    if fixed_size + final_length > 20:
        raise RPCInvalidReturnValueError(0, 0, code, resp, reason="Variable length return value is too large for rpc response payload (20 bytes)",
                                         fixed_code=code, fixed_length=fixed_size, variable_length=final_length)

    return "<" + code + "%ds" % final_length


def pack_rpc_response(response=None, exception=None):
    """Convert a response payload or exception to a status code and payload.

    This function will convert an Exception raised by an RPC implementation
    to the corresponding status code.
    """

    if response is None:
        response = bytes()

    if exception is None:
        status = (1 << 6)
        if len(response) > 0:
            status |= (1 << 7)
    elif isinstance(exception, (RPCInvalidIDError, RPCNotFoundError)):
        status = 2
    elif isinstance(exception, BusyRPCResponse):
        status = 0
    elif isinstance(exception, TileNotFoundError):
        status = 0xFF
    elif isinstance(exception, RPCErrorCode):
        status = (1 << 6) | (exception.params['code'] & ((1 << 6) - 1))
    else:
        status = 3

    return status, response


def unpack_rpc_response(status, response=None, rpc_id=0, address=0):
    """Unpack an RPC status back in to payload or exception."""

    status_code = status & ((1 << 6) - 1)

    if address == 8:
        status_code &= ~(1 << 7)

    # There is a firmware bug in lib_controller that misreports rpc not found exceptions
    # as application level exceptions, not protocol level exceptions.
    if status == 0:
        raise BusyRPCResponse()
    elif status == 2 or (address == 8 and status_code == 2):
        raise RPCNotFoundError("rpc %d:%04X not found" % (address, rpc_id))
    elif status == 3:
        raise RPCErrorCode(status_code)
    elif status == 0xFF:
        raise TileNotFoundError("tile %d not found" % address)
    elif status_code != 0:
        raise RPCErrorCode(status_code)

    if response is None:
        response = b''

    return response


def pack_rpc_payload(arg_format, args):
    """Pack an RPC payload according to arg_format.

    Args:
        arg_format (str): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        args (list): A list of arguments to pack according to arg_format.

    Returns:
        bytes: The packed argument buffer.
    """

    code = _create_respcode(arg_format, args)

    packed_result = struct.pack(code, *args)
    unpacked_validation = struct.unpack(code, packed_result)
    if tuple(args) != unpacked_validation:
        raise RPCInvalidArgumentsError("Passed values would be truncated, please validate the size of your string",
                                       code=code, args=args)
    return packed_result


def unpack_rpc_payload(resp_format, payload):
    """Unpack an RPC payload according to resp_format.

    Args:
        resp_format (str): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        payload (bytes): The binary payload that should be unpacked.

    Returns:
        list: A list of the unpacked payload items.
    """

    code = _create_argcode(resp_format, payload)
    return struct.unpack(code, payload)


def rpc(address, rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID and address.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        address (int): The address of the mock tile this RPC is for
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC. This format code may include the final
            character V, which means that it expects a variable length bytearray.
    """

    if rpc_id < 0 or rpc_id > 0xFFFF:
        raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

    def _rpc_wrapper(func):
        async def _rpc_executor(self, payload):
            try:
                args = unpack_rpc_payload(arg_format, payload)
            except struct.error as exc:
                raise RPCInvalidArgumentsError(str(exc), arg_format=arg_format, payload=binascii.hexlify(payload))

            resp = func(self, *args)
            if inspect.isawaitable(resp):
                resp = await resp

            if resp is None:
                resp = []

            if resp_format is not None:
                try:
                    return pack_rpc_payload(resp_format, resp)
                except struct.error as exc:
                    raise RPCInvalidReturnValueError(address, rpc_id, resp_format, resp, error=exc) from exc

            return resp

        _rpc_executor.rpc_id = rpc_id
        _rpc_executor.rpc_addr = address
        _rpc_executor.is_rpc = True
        return _rpc_executor

    return _rpc_wrapper


def tile_rpc(rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID on a tile.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    return rpc(None, rpc_id, arg_format, resp_format)


class RPCDispatcher:
    """A simple dispatcher that can store and call RPCs."""

    def __init__(self, *args, **kwargs):
        super(RPCDispatcher, self).__init__(*args, **kwargs)
        self._rpcs = {}

        # Add any RPCs defined using decorators on this class
        for _name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.add_rpc(value.rpc_id, value)

    def add_rpc(self, rpc_id, func):
        """Add an RPC.

        Args:
            rpc_id (int): The ID of the RPC
            func (callable): The RPC implementation.
                The signature of callable should be callable(args) taking
                a bytes object with the argument and returning a bytes object
                with the response.
        """

        self._rpcs[rpc_id] = func

    def has_rpc(self, rpc_id):
        """Check if an RPC is defined.

        Args:
            rpc_id (int): The RPC to check

        Returns:
            bool: Whether it exists
        """

        return rpc_id in self._rpcs

    def call_rpc(self, rpc_id, payload=bytes()):
        """Call an RPC by its ID.

        Args:
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            bytes: The response payload from the RPC
        """
        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if rpc_id not in self._rpcs:
            raise RPCNotFoundError("rpc_id: {}".format(rpc_id))

        return self._rpcs[rpc_id](payload)
